-----------------------------------------------------------------------------------------------
-- Client Lua Script for Clairvoyance
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Apollo"
require "Window"
require "Unit"
require "GameLib"
require "CombatFloater"
require "math"
 
-----------------------------------------------------------------------------------------------
-- Clairvoyance Module Definition
-----------------------------------------------------------------------------------------------
local Clairvoyance = {} 
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function Clairvoyance:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

    return o
end

function Clairvoyance:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 
function Clairvoyance:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("Clairvoyance.xml")
	Apollo.LoadSprites("ClairvoyanceSprites.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

function Clairvoyance:OnDocLoaded()
	if self.xmlDoc == nil then
		return
	end
	
	Apollo.RegisterEventHandler("ApplyCCState", "OnCCApplied", self)
	Apollo.RegisterEventHandler("RemoveCCState", "OnCCRemove", self)
	Apollo.RegisterEventHandler("ChangeWorld", "OnChangeWorld", self)
	Apollo.RegisterEventHandler("UnitCreated", "OnUnitCreated", self)
	Apollo.RegisterEventHandler("SystemKeyDown", "OnKeyDown", self)
	
	self.wndClairvoyanceNotification = Apollo.LoadForm(self.xmlDoc, "ClairvoyanceNotification", nil, self)
	self.wndClairvoyanceNotification:Show(false)
	self.wndWeaponIndicator = Apollo.LoadForm(self.xmlDoc, "WeaponIndicator", nil, self)
	self.wndWeaponIndicator:Show(false)
	self.wndDisorient = Apollo.LoadForm(self.xmlDoc, "DisorientWindow", nil, self)
	self.wndDisorient:FindChild("forwardButton"):Enable(false)
	self.wndDisorient:FindChild("backButton"):Enable(false)
	self.wndDisorient:FindChild("strafeLeftButton"):Enable(false)
	self.wndDisorient:FindChild("strafeRightButton"):Enable(false)
	self.wndDisorient:Show(false)
	
	Apollo.RegisterSlashCommand("cv", "OnSlashCommand", self)
	Apollo.RegisterSlashCommand("clairvoyance", "OnSlashCommand", self)
	
	if self.displayBackground == nil then
		self.displayBackground = true
	end
	if self.savedLoc then
		self.wndClairvoyanceNotification:MoveToLocation(self.savedLoc)
	else
		self.savedLoc = self.wndClairvoyanceNotification:GetLocation()
	end
	if self.savedDisorientLoc then
		self.wndDisorient:MoveToLocation(self.savedDisorientLoc)
	else
		self.savedDisorientLoc = self.wndDisorient:GetLocation()
	end
	self.activeEffects = {}
	self:InitializeStateNames()
	self:InitializeStatePriority()
	self:InitializeKeyMapping()
end

function Clairvoyance:OnSlashCommand(slashcommand, subcommand)
	if subcommand == "background" or subcommand == "b" then
		if self.displayBackground == true then
			self.displayBackground = false
			Print("Clairvoyance - Background turned OFF for notification window.")
		elseif self.displayBackground == false then
			self.displayBackground = true
			Print("Clairvoyance - Background turned ON for notification window.")
		end
	end
end

-----------------------------------------------------------------------------------------------
-- Functionality
-----------------------------------------------------------------------------------------------

function Clairvoyance:OnCCApplied(code, userdata)
	local unitPlayer = GameLib.GetControlledUnit()
	if code == Unit.CodeEnumCCState.Stun or code == Unit.CodeEnumCCState.Knockback then
		return
	end
	if unitPlayer == userdata then
		if code == Unit.CodeEnumCCState.Disorient then
			self.originalKeybindings = GameLib.GetKeyBindings()
			self.modifiedKeybindings = GameLib.GetKeyBindings()
		end
		self.activeEffects[code] = true
		self:UpdateNotification()
	end
end

function Clairvoyance:OnCCRemove(code, userdata)
	local unitPlayer = GameLib.GetControlledUnit()
	if code == Unit.CodeEnumCCState.Stun or code == Unit.CodeEnumCCState.Knockback then
		return
	end
	if unitPlayer == userdata then
		self.activeEffects[code] = nil
		self:UpdateNotification()
		if code == Unit.CodeEnumCCState.Subdue or code == Unit.CodeEnumCCState.Disarm then
			self.wndWeaponIndicator:Show(false)
		end
		if code == Unit.CodeEnumCCState.Disorient then
			GameLib.SetKeyBindings(self.originalKeybindings)
			self.lastLocation = nil
			self.wndDisorient:Show(false)
			self.wndDisorient:FindChild("forwardButton"):DestroyAllPixies()
			self.wndDisorient:FindChild("backButton"):DestroyAllPixies()
			self.wndDisorient:FindChild("strafeLeftButton"):DestroyAllPixies()
			self.wndDisorient:FindChild("strafeRightButton"):DestroyAllPixies()
		end
	end
end

function Clairvoyance:UpdateNotification()
	local displayEffect = -1
	for key, value in pairs(self.activeEffects) do
		if (displayEffect == -1) or (self.statePriority[key] > self.statePriority[displayEffect]) then
			displayEffect = key
		end
	end
	
	if displayEffect == -1 then
		self:ShowNotification(false)
	else
		self:ShowNotification(false)
		
		if displayEffect == Unit.CodeEnumCCState.Knockdown then
			self.wndClairvoyanceNotification:FindChild("Advice"):SetText("Dodge to recover!")
		elseif displayEffect == Unit.CodeEnumCCState.Disorient then
			self.wndClairvoyanceNotification:FindChild("Advice"):SetText("Movement keys changed")
		elseif displayEffect == Unit.CodeEnumCCState.Subdue or displayEffect == Unit.CodeEnumCCState.Disarm then
			self.wndClairvoyanceNotification:FindChild("Advice"):SetText("Find your weapon!")
		elseif displayEffect == Unit.CodeEnumCCState.Tether then
			self.wndClairvoyanceNotification:FindChild("Advice"):SetText("Destroy the anchor!")
		else
			self.wndClairvoyanceNotification:FindChild("Advice"):SetText("")
		end
		if self.stateNames[displayEffect] ~= nil then
			self.wndClairvoyanceNotification:FindChild("StateName"):SetText(self.stateNames[displayEffect])
		else
			Print("Received code " .. displayEffect .. "...")
			self.wndClairvoyanceNotification:FindChild("StateName"):SetText("Undefined CC")
		end
		if self.displayBackground == true then
			self.wndClairvoyanceNotification:SetSprite("BK3:UI_BK3_Holo_Framing_2_Blocker")
		elseif self.displayBackground == false then
			self.wndClairvoyanceNotification:SetSprite("")
		end
		
		self:ShowNotification(true)
	end
end

function Clairvoyance:OnChangeWorld()
	for key, value in pairs(self.activeEffects) do
		self.activeEffects[key] = nil
	end
	self:ShowNotification(false)
	self.wndWeaponIndicator:Show(false)
end

function Clairvoyance:OnUnitCreated(unit)
	if unit:GetType() == "Pickup" then
		local playerName = GameLib.GetPlayerUnit():GetName()
		if string.sub(unit:GetName(), 1, string.len(playerName)) == playerName then
			self.wndWeaponIndicator:SetUnit(unit, 1)
			self.wndWeaponIndicator:Show(true)
		end
	end
end

function Clairvoyance:OnKeyDown(iKey)
	if self.activeEffects[Unit.CodeEnumCCState.Disorient] == true then
		if self.keymapping ~= nil and self.keymapping[iKey] ~= nil then
			local key = self.keymapping[iKey]
			local strafeLeft = GameLib.GetKeyBinding("StrafeLeft")
			local strafeRight = GameLib.GetKeyBinding("StrafeRight")
			local moveForward = GameLib.GetKeyBinding("MoveForward")
			local moveBackward = GameLib.GetKeyBinding("MoveBackward")
			if key == strafeLeft or key == strafeRight or key == moveForward or key == moveBackward then
				self.wndDisorient:Show(true)
				self:AttemptRebind(key, strafeLeft, strafeRight, moveForward, moveBackward)
			end
		end
	end
end

function Clairvoyance:AttemptRebind(direction, strafeLeft, strafeRight, moveForward, moveBackward)
	local unitPlayer = GameLib.GetControlledUnit()
	if self.lastLocation == nil or self.lastDirection == nil then
		self.lastLocation = unitPlayer:GetPosition()
		self.lastDirection = direction
	else
		if self.lastDirection ~= direction then
			self.lastDirection = direction
		else
			self.lastDirection = direction
			local currentPosition = unitPlayer:GetPosition()
			local actualDirection = self:CalculateDirection(self.lastLocation, currentPosition, unitPlayer:GetHeading())
			self.lastLocation = currentPosition
			if direction == moveForward then
				if actualDirection == "FORWARD" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("forwardButton"), 0)
				elseif actualDirection == "BACKWARD" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("forwardButton"), 180)
				elseif actualDirection == "LEFTSTRAFE" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("forwardButton"), 270)
				elseif actualDirection == "RIGHTSTRAFE" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("forwardButton"), 90)
				end
			elseif direction == moveBackward then
				if actualDirection == "FORWARD" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("backButton"), 0)
				elseif actualDirection == "BACKWARD" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("backButton"), 180)
				elseif actualDirection == "LEFTSTRAFE" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("backButton"), 270)
				elseif actualDirection == "RIGHTSTRAFE" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("backButton"), 90)
				end
			elseif direction == strafeLeft then
				if actualDirection == "FORWARD" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("strafeLeftButton"), 0)
				elseif actualDirection == "BACKWARD" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("strafeLeftButton"), 180)
				elseif actualDirection == "LEFTSTRAFE" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("strafeLeftButton"), 270)
				elseif actualDirection == "RIGHTSTRAFE" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("strafeLeftButton"), 90)
				end
			elseif direction == strafeRight then
				if actualDirection == "FORWARD" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("strafeRightButton"), 0)
				elseif actualDirection == "BACKWARD" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("strafeRightButton"), 180)
				elseif actualDirection == "LEFTSTRAFE" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("strafeRightButton"), 270)
				elseif actualDirection == "RIGHTSTRAFE" then
					self:DrawArrowPixie(self.wndDisorient:FindChild("strafeRightButton"), 90)
				end
			end
		end
	end
end

function Clairvoyance:CalculateDirection(fromLoc, toLoc, heading)
	local angle = self:ComputeAngle(fromLoc, toLoc)
	local backward = self:FindBackward(heading)
	local leftStrafe = self:FindLeftStrafe(heading)
	local rightStrafe = self:FindRightStrafe(heading)
	
	local match = self:CompareAngles(angle, heading, 0.2)
	if match == true then
		return "FORWARD"
	end
	match = self:CompareAngles(angle, backward, 0.2)
	if match == true then
		return "BACKWARD"
	end
	match = self:CompareAngles(angle, leftStrafe, 0.2)
	if match == true then
		return "LEFTSTRAFE"
	end
	match = self:CompareAngles(angle, rightStrafe, 0.2)
	if match == true then
		return "RIGHTSTRAFE"
	end
	
	return "?"
end

function Clairvoyance:ComputeAngle(fromLoc, toLoc)
	local dX = toLoc.x - fromLoc.x
	local dZ = toLoc.z - fromLoc.z
	local radians = -math.atan2(dZ, dX)
	if radians >= -(math.pi / 2) then
		radians = radians - (math.pi / 2)
	elseif radians < -(math.pi / 2) then
		local remainder = -radians - (math.pi / 2)
		radians = math.pi - remainder
	end
	return radians
end

function Clairvoyance:FindBackward(angle)
	local convertedAngle = nil
	if angle == 0 then
		convertedAngle = math.pi
	elseif angle == math.pi then
		convertedAngle = 0
	elseif angle > 0 then
		local remainder = math.pi - angle
		convertedAngle = 0 - remainder
	elseif angle < 0 then
		local remainder = -math.pi - angle
		convertedAngle = 0 - remainder
	end
	return convertedAngle
end

function Clairvoyance:FindLeftStrafe(angle)
	local convertedAngle = nil
	if angle > (math.pi / 2) then
		local remainder = angle - (math.pi / 2)
		convertedAngle = -math.pi + remainder
	else
		convertedAngle = angle + (math.pi / 2)
	end
	return convertedAngle
end

function Clairvoyance:FindRightStrafe(angle)
	local convertedAngle = nil
	if angle < (-math.pi / 2) then
		local remainder = angle - (-math.pi / 2)
		convertedAngle = math.pi + remainder
	else
		convertedAngle = angle - (math.pi / 2)
	end
	return convertedAngle
end

function Clairvoyance:CompareAngles(angle, heading, faultTolerance)
	local min = nil
	local max = nil
	min, max = self:FindBoundaries(heading, faultTolerance)
	if min > max then
		if (angle >= min and angle <= math.pi) or (angle <= max and angle >= -math.pi) then
			return true
		end
	else
		if angle >= min and angle <= max then
			return true
		end
	end
	
	return false
end

function Clairvoyance:FindBoundaries(angle, faultTolerance)
	local min = nil
	local max = nil
	if (angle + faultTolerance) > math.pi then
		min = angle - faultTolerance
		local remainder = math.pi - angle
		remainder = faultTolerance - remainder
		max = -math.pi + remainder
	elseif (angle - faultTolerance) < -math.pi then
		max = angle + faultTolerance
		local remainder = -math.pi - angle
		remainder = faultTolerance - remainder
		min = math.pi - remainder
	else
		min = angle - faultTolerance
		max = angle + faultTolerance
	end
	return min, max
end

function Clairvoyance:round(num, idp)
	local mult = 10^(idp or 0)
	return math.floor(num * mult + 0.5) / mult
end

function Clairvoyance:ShowNotification(show)
	self.wndClairvoyanceNotification:Show(show)
	self.wndClairvoyanceNotification:FindChild("Icon"):Show(show)
	self.wndClairvoyanceNotification:FindChild("StateName"):Show(show)
	self.wndClairvoyanceNotification:FindChild("Advice"):Show(show)
end

function Clairvoyance:DrawArrowPixie(component, rotation)
	component:DestroyAllPixies()
	
	local loc = {
		fPoints = {0.5, 0.5, 0.5, 0.5},
		nOffsets = {-14, -14, 14, 14}
	}
	
	component:AddPixie({
	iLayer = 2,
	bLine = false,
	strSprite = "ClairvoyanceSprites:Arrow",
	fRotation = rotation,
	loc = loc})
end

function Clairvoyance:InitializeStateNames()
	self.stateNames = {}
	for key, value in pairs(Unit.CodeEnumCCState) do
		self.stateNames[value] = key
	end
end

function Clairvoyance:InitializeStatePriority()
	self.statePriority = {}
	for key, value in pairs(Unit.CodeEnumCCState) do
		self.statePriority[value] = 1
	end
	self.statePriority[Unit.CodeEnumCCState.Knockdown] = 6
	self.statePriority[Unit.CodeEnumCCState.Subdue] = 5
	self.statePriority[Unit.CodeEnumCCState.Disarm] = 4
	self.statePriority[Unit.CodeEnumCCState.Tether] = 3
	self.statePriority[Unit.CodeEnumCCState.Disorient] = 2
end

function Clairvoyance:InitializeKeyMapping()
	self.keymapping = {}
	self.keymapping[81] = "Q"
	self.keymapping[87] = "W"
	self.keymapping[69] = "E"
	self.keymapping[82] = "R"
	self.keymapping[84] = "T"
	self.keymapping[89] = "Y"
	self.keymapping[85] = "U"
	self.keymapping[73] = "I"
	self.keymapping[79] = "O"
	self.keymapping[80] = "P"
	self.keymapping[65] = "A"
	self.keymapping[83] = "S"
	self.keymapping[68] = "D"
	self.keymapping[70] = "F"
	self.keymapping[71] = "G"
	self.keymapping[72] = "H"
	self.keymapping[74] = "J"
	self.keymapping[75] = "K"
	self.keymapping[76] = "L"
	self.keymapping[90] = "Z"
	self.keymapping[88] = "X"
	self.keymapping[67] = "C"
	self.keymapping[86] = "V"
	self.keymapping[66] = "B"
	self.keymapping[78] = "N"
	self.keymapping[77] = "M"
end

---------------------------------------------------------------------------------------------------
-- Options
---------------------------------------------------------------------------------------------------

function Clairvoyance:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return nil
	end

	local loc = self.wndClairvoyanceNotification and self.wndClairvoyanceNotification:GetLocation() or self.savedLoc
	local disorientLoc = self.wndDisorient and self.wndDisorient:GetLocation() or self.savedDisorientLoc
	
	local tSaved =
	{
		tLocation = loc and loc:ToTable() or nil,
		tDisplayBackground = self.displayBackground,
		tDisorientLocation = disorientLoc and disorientLoc:ToTable() or nil,
	}

	return tSaved
end

function Clairvoyance:OnRestore(eType, tSavedData)
	if eType == GameLib.CodeEnumAddonSaveLevel.Character and tSavedData then
		if tSavedData.tLocation ~= nil then
			self.savedLoc = WindowLocation.new(tSavedData.tLocation)
		end
		if tSavedData.tDisplayBackground ~= nil then
			self.displayBackground = tSavedData.tDisplayBackground
		end
		if tSavedData.tDisorientLocation ~= nil then
			self.savedDisorientLoc = WindowLocation.new(tSavedData.tDisorientLocation)
		end
	end
end

-----------------------------------------------------------------------------------------------
-- Clairvoyance Instance
-----------------------------------------------------------------------------------------------
local ClairvoyanceInst = Clairvoyance:new()
ClairvoyanceInst:Init()
